#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "globals.h"
#include "CElem.hpp"
#include "paint.h"
#include "algebra.h"

using namespace std;

unsigned CElem::elemcnt = 0;

CElem::CElem()
 {
	++elemcnt;
	points = NULL;
	common.pointcnt = 0;
	common.depth = 0;
	common.selected = false;
	common.visible = true;
	common.groupId = 0;   // single element group
	common.lineWidth = 1;
	common.lineStyle = DASH_SOLID;
	common.filled = 0;
 }

CElem::~CElem()
 {
	--elemcnt;
 }

void CElem::addPoint( unsigned pt, double newX, double newY )
 {
	struct tPointLink *act = points, *prev = NULL, *temp;
	unsigned ord = 0;

	while( act != NULL && ord < pt )
	 {
		prev = act;
		act = act->next;
		++ord;
	 }
	temp = new tPointLink;
	if( temp == NULL ) error( ERROR_NEW, __FILE__, __LINE__ );
	temp->x = newX;
	temp->y = newY;
	if( prev != NULL )
	 {
		temp->next = act;
		prev->next = temp;
	 }
	else   // we are inserting at the beginning of list
	 {
		temp->next = points;
		points = temp;
	 }
	++common.pointcnt;
	return;
 }

bool CElem::delPoint( unsigned pt )
 {
	struct tPointLink *act = points, *prev = NULL;
	unsigned ord = 0;

	while( act != NULL && ord < pt )
	 {
		prev = act;
		act = act->next;
		++ord;
	 }
	if( prev != NULL ) prev->next = act->next;
	else   // we are deleting from the beginning of list
		points = act->next;
	if( act == NULL ) return false;
	delete act;
	--common.pointcnt;
	return true;
 }

bool CElem::movePoint( unsigned pt, double newX, double newY )
 {
	struct tPointLink *act = points;
	unsigned ord = 0;

	while( act != NULL && ord < pt )
	 {
		act = act->next;
		++ord;
	 }
	if( act == NULL ) return false;
	act->x = newX;
	act->y = newY;
	return true;
 }

bool CElem::getPoint( unsigned pt, double *oldX, double *oldY )
 {
	struct tPointLink *act = points;
	unsigned ord = 0;

	while( act != NULL && ord < pt )
	 {
		act = act->next;
		++ord;
	 }
	if( act == NULL ) return false;
	*oldX = act->x;
	*oldY = act->y;
	return true;
 }

void *CElem::save( int *len ) const
 {
	void *res;
	unsigned offset = 0;

	/* here we save at first type of element (int), then common atributes and last all points */
	/* for each point, tPointLink structure is saved, except of pointer to the next (linked list) */
	res = malloc( sizeof(int) + sizeof(common) + common.pointcnt*(sizeof(*points) - sizeof(points)) );
	if ( res == NULL ) error( ERROR_MALLOC, __FILE__, __LINE__ );

	*(int *)(res) = elemtype;
	offset += sizeof(elemtype);
	saveCommon( (void *)((char *)res + offset) );
	offset += sizeof(common);
	/* saving linked list... */
	struct tPointLink *act = points;
	while( act != NULL )
	 {
		memcpy( (void *)( (char *)res + offset), act, (sizeof(*points) - sizeof(points)) );
		offset += (sizeof(*points) - sizeof(points));
	 }
	*len = offset;
	return res;
 }

int CElem::load( const void *ptr )
 {
	//TODO
	return 0;
 }

void CElem::allocPoints()
 {
	unsigned rem = common.pointcnt;
	tPointLink *temp;

	deallocPoints();   // pointcnt must be consistent with length of list!
	while( rem-- )
	 {
		temp = new tPointLink;
		if( temp == NULL ) error( ERROR_NEW, __FILE__, __LINE__ );
		temp->next = points;
		points = temp;
	 }
	return;
 }

void CElem::deallocPoints()
 {
	tPointLink *temp;

	while( points != NULL )
	 {
		temp = points;
		points = points->next;
		delete temp;
	 }
	return;
 }

void CElem::saveCommon( void *ptr ) const
 {
	memcpy( ptr, &common, sizeof(common) );
	return;
 }

void CElem::loadCommon( const void *ptr )
 {
	memcpy( &common, ptr, sizeof(common) );
	return;
 }

void CElem::drawElems( vector<CElem *> &elems )
 {
	CElem::tPoint *box;
	unsigned boxpoints;
	vector<CElem *>::iterator iter = elems.begin();

	while( iter != elems.end() )
	 {
		if( !(*iter)->common.visible ) {iter++; continue;}
		(*iter)->draw();
		if( !(*iter)->common.selected ) {iter++; continue;}
		box = (*iter)->boundary( &boxpoints );
		if( box == NULL ) {iter++; continue;}   // empty elements return NULL instead of boundary
		if( (*iter)->common.groupId == ELEM_GROUP_SINGLE )
		 {
			// this cast is potentially dangerous!, but for this time we will believe that struct point_s (painting) and tPoint (CElem) are the same
			draw_hilighted_points( (struct point_s *)box, boxpoints );
		 }
		else   // we are drawing element from group => determining group boundary
		 {
			//TODO: tady se budou kreslit boxy cele grupy a ne jednotlivych elementu
			draw_hilighted_grouppoints( (struct point_s *)box, boxpoints );
		 }
		delete [] box;
		iter++;	
	 }
	return;
 }

int CElem::selectElems( std::vector<CElem *> &elems, double x, double y, bool additive, double tolerance )
 {
	vector<CElem *>::iterator iter = elems.begin(); 
	int res = 0;
	struct tPoint click = {x, y};

	// at first unselected uninteresting elements
	while( iter != elems.end() )
	 {
		if( additive ) break;   // if additive is active, keep old selection
		(*iter)->common.selected = false;
		iter++;
	 }
	// now select iteresting ones...
	iter = elems.begin();
	while( iter != elems.end() )
	 {
		if( !(*iter)->common.visible ) {iter++; continue;}
		if( (*iter)->distance( click ) < tolerance )
		 {
			(*iter)->common.selected = true;
			if( (*iter)->common.groupId != ELEM_GROUP_SINGLE )
			 {   // ...with regard to groups!
				vector<CElem *>::iterator itergroup = elems.begin();	
				while( itergroup != elems.end() ) {
					if( (*itergroup)->common.groupId == (*iter)->common.groupId )
					 {
						(*itergroup)->common.selected = true;
						++res;
					 }
					itergroup++;
				}
			 }
			else ++res;   // alone element => increment by one
		 }
		iter++;	
	 }
	return res;
 }

int CElem::selectElemsInside( std::vector<CElem *> &elems, double x1, double y1, double x2, double y2, bool additive, bool whole )
 {
	vector<CElem *>::iterator iter = elems.begin(); 
	int res = 0;
	struct tPoint pt1 = {x1, y1};
	struct tPoint pt2 = {x2, y2};

	// at first unselected uninteresting elements
	while( iter != elems.end() )
	 {
		if( additive ) break;   // if additive is active, keep old selection
		(*iter)->common.selected = false;
		iter++;
	 }
	// now select iteresting ones...
	iter = elems.begin();
	while( iter != elems.end() )
	 {
		if( !(*iter)->common.visible ) {iter++; continue;}
		if( (*iter)->inside( pt1, pt2, whole ) )
		 {
			(*iter)->common.selected = true;
			if( (*iter)->common.groupId != ELEM_GROUP_SINGLE )
			 {   // ...with regard to groups!
				//TODO: tady by se spravne mely selektovat grupy v zavislosti na whole jenom tehdy jsou-li ve vyberu vsechny body vsech prvku grupy, nikoliv ze staci jeden
				vector<CElem *>::iterator itergroup = elems.begin();	
				while( itergroup != elems.end() ) {
					if( (*itergroup)->common.groupId == (*iter)->common.groupId )
					 {
						(*itergroup)->common.selected = true;
						++res;
					 }
					itergroup++;
				}
			 }
			else ++res;   // alone element => increment by one
		 }
		iter++;	
	 }
	return res;
 }

int CElem::groupElems( std::vector<CElem *> &elems )
 {
	vector<CElem *>::iterator iter = elems.begin(); 
	CElem *lastiter;
	int lastGroupId;
	int res = 0;

	while( iter != elems.end() )
	 {
		if( lastGroupId < (*iter)->common.groupId ) lastGroupId = (*iter)->common.groupId;
		iter++;	
	 }
//TODO: tady se musi resit sklepavani indexu groupId pokud jsou obsazene az do nejvyssich cisel
	iter = elems.begin();
	while( iter != elems.end() )
	 {
		if( (*iter)->common.selected )
		 {
			(*iter)->common.groupId = lastGroupId+1;
			lastiter = (*iter);
			++res;
		 }
		iter++;
	 }
	if( res == 1 ) // if only one selected, it will not form group of one element
	 {
		lastiter->common.groupId = ELEM_GROUP_SINGLE;
		return 0;
	 }
	return res;
 }

int CElem::ungroupElems( std::vector<CElem *> &elems )
 {
	vector<CElem *>::iterator iter = elems.begin(); 
	int res = 0;

	while( iter != elems.end() )
	 {
		if( (*iter)->common.selected )
		 {
			(*iter)->common.groupId = ELEM_GROUP_SINGLE;
			++res;
		 }
		iter++;	
	 }
	return res;
 }

int CElem::moveElems( std::vector<CElem *> &elems, double x, double y )
 {
	vector<CElem *>::iterator iter = elems.begin(); 
	int res = 0;

	while( iter != elems.end() )
	 {
		if( (*iter)->common.selected )
		 {
			struct tPointLink *act = (*iter)->points;
			unsigned ord = 0;

			while( act != NULL && ord < (*iter)->common.pointcnt )
			 {
				act->x += x;
				act->y += y;
				act = act->next;
				++ord;
			 }
			++res;
		 }
		iter++;	
	 }
	return res;
 }

int CElem::delElems( std::vector<CElem *> &elems )
 {
	vector<CElem *>::iterator iter = elems.begin();
	int res = 0;

	while ( iter != elems.end() )
	 {
		if( *iter == NULL ) {iter++; continue;}
		if( (*iter)->common.selected )
		 {
			delete (*iter);
			elems.erase( iter );
			++res;
		 }
		else iter++;
	 }
	return res;
 }

CElemLines::CElemLines( unsigned pointcnt ) 
 {
	elemtype = ELEM_TYPE_LINES;
	common.pointcnt = pointcnt; /*linecnt+1;*/
	if( common.pointcnt ) allocPoints();
 }

CElemLines::~CElemLines()
 {
	deallocPoints();
 }

void CElemLines::draw() const
 {
	struct tPointLink *ptin = points;
	struct point_s *ptout;
	unsigned i = 0; 

	ptin = points;
	ptout = (struct point_s *)malloc( common.pointcnt*sizeof(*ptout) );
	if ( ptout == NULL ) error( ERROR_MALLOC, __FILE__, __LINE__ );
	while( i < common.pointcnt  )
	 {
		if( ptin == NULL ) break;   // paranoia
		ptout[i].x = ptin->x;
		ptout[i].y = ptin->y;
		ptin = ptin->next;
		++i;
	 }
	if( i > 0 )   // paranoia
		draw_multiline( ptout, i, 
		                common.colorR, common.colorG, common.colorB,
		                common.lineWidth, common.lineStyle );
	return;
 }

struct CElem::tPoint *CElemLines::boundary( unsigned *ptcnt ) const
 {
	struct tPointLink *act = points;
	unsigned ord = 0;
	struct CElem::tPoint *res;

	// if there are no defined points yet, no boundary will be generated
	if( !common.pointcnt ) return (struct tPoint *)NULL;
	*ptcnt = common.pointcnt;   // all inner points
	res = new CElem::tPoint[*ptcnt];
	if( res == NULL ) error( ERROR_NEW, __FILE__, __LINE__ );
	// testing ord here is array-bounds paranoia (what if list length > pointcnt due to bug somewhere else)
	while( act != NULL && ord < common.pointcnt )
	 {
		res[ord].x = act->x;
		res[ord].y = act->y;
		act = act->next;
		++ord;
	 }
	return res;
 }

double CElemLines::distance( struct CElem::tPoint click ) const
 {
	struct tPointLink *act = points, *prev = NULL;
	double res = INFINITY;

	while( act != NULL )
	 {
		//if( prev != NULL )
		 { //TODO: tady se budou resit vzdalenosti od usecek a ne jenom vsech bodu, jejich krajni body jsou v act a prev
			double dist = (click.x-act->x)*(click.x-act->x) + (click.y-act->y)*(click.y-act->y);
			if( dist < res ) res = dist; 
		 }
		prev = act;
		act = act->next;
	 }
	return res;
 }

bool CElemLines::inside( struct CElem::tPoint pt1, struct CElem::tPoint pt2, bool whole ) const
 {
	struct tPointLink *act = points;

	if( act == NULL ) return false;
	while( act != NULL )
	 {
		if( whole )
		 {
			if( !betweend( pt1.x, act->x, pt2.x ) || !betweend( pt1.y, act->y, pt2.y ) )
				return false;
		 }
		else   // one point inside is enough
		 {
			if( betweend( pt1.x, act->x, pt2.x ) && betweend( pt1.y, act->y, pt2.y ) )
				return true;
		 }
		act = act->next;
	 }
	if( whole ) return true;   // no point is outside
	return false;              // no point is inside
 }

CElemEllipse::CElemEllipse()
 {
	elemtype = ELEM_TYPE_ELLIPSE;
  /*
	common.pointcnt = 2;   // center point & right down on boundary box
	if( common.pointcnt ) allocPoints();
  */
 }

CElemEllipse::~CElemEllipse()
 {
	deallocPoints();
 }

void CElemEllipse::draw() const
 {
	double cntx, cnty;
	double radh, radv;
	
	if( points != NULL )
	 {
		if( points->next != NULL )
		 {
			cntx = points->x;
			cnty = points->y;
			radh = points->next->x - cntx;
			radv = points->next->y - cnty;

			draw_ellipse( cntx, cnty, radh, radv,
			              common.colorRbg, common.colorGbg, common.colorBbg,
			              common.colorR, common.colorG, common.colorB, 
			              common.lineWidth, common.lineStyle, common.filled );
		 }
	 }
			
	return;
 }

struct CElem::tPoint *CElemEllipse::boundary( unsigned *ptcnt ) const
 {
	struct CElem::tPoint *res;
	double cntx, cnty;
	double radh, radv;

	// if there are no defined points yet, no boundary will be generated
	if( points == NULL ) return (struct tPoint *)NULL;
	// if there is only center, boundary will be only this point
	if( points->next == NULL ) *ptcnt = 1;
	else *ptcnt = 9;   // center point and 8 around 
	res = new CElem::tPoint[*ptcnt];
	if( res == NULL ) error( ERROR_NEW, __FILE__, __LINE__ );
	cntx = points->x;
	cnty = points->y;
	res[0].x = cntx; res[0].y = cnty;
	if( *ptcnt == 1 ) return res;
	radh = points->next->x - cntx;
	radv = points->next->y - cnty; 
	res[1].x = cntx-radh; res[1].y = cnty-radv;
	res[2].x = cntx     ; res[2].y = cnty-radv;
	res[3].x = cntx+radh; res[3].y = cnty-radv;
	res[4].x = cntx+radh; res[4].y = cnty     ;
	res[5].x = cntx+radh; res[5].y = cnty+radv;
	res[6].x = cntx     ; res[6].y = cnty+radv;
	res[7].x = cntx-radh; res[7].y = cnty+radv;
	res[8].x = cntx-radh; res[8].y = cnty     ;
	return res;
 }

double CElemEllipse::distance( struct CElem::tPoint click ) const
 {
	double res = INFINITY;
	double cntx, cnty;
	double radh, radv;

	if( points == NULL ) return res;
	cntx = points->x;
	cnty = points->y;
	res = (click.x-cntx)*(click.x-cntx) + (click.y-cnty)*(click.y-cnty);
	// paranoia (ellipse should not be selectable by user when it has defined only center point)
	if( points->next == NULL ) return res;
	radh = points->next->x - cntx;
	radv = points->next->y - cnty; 
	//TODO:pocitat vzdalenost primo od cary elipsy?
	// get distances between click and intersection points of ellipse and XYaxes 
	res = min( res, (click.x-cntx-radh)*(click.x-cntx-radh) + (click.y-cnty)*(click.y-cnty) );
	res = min( res, (click.x-cntx+radh)*(click.x-cntx+radh) + (click.y-cnty)*(click.y-cnty) );
	res = min( res, (click.x-cntx)*(click.x-cntx) + (click.y-cnty-radv)*(click.y-cnty-radv) );
	res = min( res, (click.x-cntx)*(click.x-cntx) + (click.y-cnty+radv)*(click.y-cnty+radv) );
	return res;
 }

bool CElemEllipse::inside( struct CElem::tPoint pt1, struct CElem::tPoint pt2, bool whole ) const
 {
	double cntx, cnty;
	double radh, radv;

	if( points == NULL ) return false;
	cntx = points->x;
	cnty = points->y;
	if(  whole ) if ( !betweend( pt1.x, cntx, pt2.x ) || !betweend( pt1.y, cnty, pt2.y ) ) return false;
	if( !whole ) if (  betweend( pt1.x, cntx, pt2.x ) &&  betweend( pt1.y, cnty, pt2.y ) ) return true;
	// paranoia (as in method distance)
	if( points->next == NULL )
	 {
		if( whole ) return true;
		return false;
	 }
	radh = points->next->x - cntx;
	radv = points->next->y - cnty; 
	// points of interest are intersection points of ellipse and XYaxes 
	if(  whole ) if ( !betweend( pt1.x, cntx-radh, pt2.x ) || !betweend( pt1.y, cnty, pt2.y ) ) return false;
	if( !whole ) if (  betweend( pt1.x, cntx-radh, pt2.x ) &&  betweend( pt1.y, cnty, pt2.y ) ) return true;
	if(  whole ) if ( !betweend( pt1.x, cntx+radh, pt2.x ) || !betweend( pt1.y, cnty, pt2.y ) ) return false;
	if( !whole ) if (  betweend( pt1.x, cntx+radh, pt2.x ) &&  betweend( pt1.y, cnty, pt2.y ) ) return true;
	if(  whole ) if ( !betweend( pt1.x, cntx, pt2.x ) || !betweend( pt1.y, cnty-radv, pt2.y ) ) return false;
	if( !whole ) if (  betweend( pt1.x, cntx, pt2.x ) &&  betweend( pt1.y, cnty-radv, pt2.y ) ) return true;
	if(  whole ) if ( !betweend( pt1.x, cntx, pt2.x ) || !betweend( pt1.y, cnty+radv, pt2.y ) ) return false;
	if( !whole ) if (  betweend( pt1.x, cntx, pt2.x ) &&  betweend( pt1.y, cnty+radv, pt2.y ) ) return true;
	if( whole ) return true;   // no point is outside
	return false;              // no point is inside
 }

CElemPolygon::CElemPolygon( unsigned linecnt ) 
 {
	elemtype = ELEM_TYPE_POLYGON;
	common.pointcnt = linecnt;
	if( common.pointcnt ) allocPoints();
 }

CElemPolygon::~CElemPolygon()
 {
	deallocPoints();
 }

void CElemPolygon::draw() const
 {
	struct tPointLink *ptin = points;
	struct point_s *ptout;
	unsigned i = 0; 

	ptin = points;
	ptout = (struct point_s *)malloc( common.pointcnt*sizeof(*ptout) );
	if ( ptout == NULL ) error( ERROR_MALLOC, __FILE__, __LINE__ );
	while( i < common.pointcnt  )
	 {
		if( ptin == NULL ) break;   // paranoia
		ptout[i].x = ptin->x;
		ptout[i].y = ptin->y;
		ptin = ptin->next;
		++i;
	 }
	if( i > 0 )   // paranoia
		draw_polygon( ptout, i, 
		              common.colorRbg, common.colorGbg, common.colorBbg,
		              common.colorR, common.colorG, common.colorB,
		              common.lineWidth, common.lineStyle, common.filled );
	return;
 }

struct CElem::tPoint *CElemPolygon::boundary( unsigned *ptcnt ) const
 {
	struct tPointLink *act = points;
	unsigned ord = 0;
	struct CElem::tPoint *res;
	double minx = INFINITY, maxx = -INFINITY;
	double miny = INFINITY, maxy = -INFINITY;

	// if there are no defined points yet, no boundary will be generated
	if( !common.pointcnt ) return (struct tPoint *)NULL;
	*ptcnt = common.pointcnt+4;   // all inner points and 4 outbox corners
	res = new CElem::tPoint[*ptcnt];
	if( res == NULL ) error( ERROR_NEW, __FILE__, __LINE__ );
	// testing ord here is array-bounds paranoia (what if list length > pointcnt due to bug somewhere else)
	while( act != NULL && ord < common.pointcnt )
	 {
		res[ord].x = act->x;
		res[ord].y = act->y;
		if( res[ord].x < minx ) minx = res[ord].x;
		if( res[ord].x > maxx ) maxx = res[ord].x;
		if( res[ord].y < miny ) miny = res[ord].y;
		if( res[ord].y > maxy ) maxy = res[ord].y;
		act = act->next;
		++ord;
	 }
	// now 4 corner points from saved extremes
	res[ord].x = minx; res[ord++].y = miny;
	res[ord].x = maxx; res[ord++].y = miny;
	res[ord].x = maxx; res[ord++].y = maxy;
	res[ord].x = minx; res[ord++].y = maxy;
	return res;
 }

double CElemPolygon::distance( struct CElem::tPoint click ) const
 {
	struct tPointLink *act = points, *prev = NULL;
	double res = INFINITY;

	while( act != NULL )
	 {
		//if( prev != NULL )
		 { //TODO: tady se budou resit vzdalenosti od usecek polygonu, ev. jestli jsme uvnitr pro filled=true
			double dist = (click.x-act->x)*(click.x-act->x) + (click.y-act->y)*(click.y-act->y);
			if( dist < res ) res = dist; 
		 }
		prev = act;
		act = act->next;
	 }
	return res;
 }

bool CElemPolygon::inside( struct CElem::tPoint pt1, struct CElem::tPoint pt2, bool whole ) const
 {
	struct tPointLink *act = points;

	if( act == NULL ) return false;
	while( act != NULL )
	 {
		if( whole )
		 {
			if( !betweend( pt1.x, act->x, pt2.x ) || !betweend( pt1.y, act->y, pt2.y ) )
				return false;
		 }
		else   // one point inside is enough
		 {
			if( betweend( pt1.x, act->x, pt2.x ) && betweend( pt1.y, act->y, pt2.y ) )
				return true;
		 }
		act = act->next;
	 }
	if( whole ) return true;   // no point is outside
	return false;              // no point is inside
 }

