#include <vector>

#ifdef DEBUG
	#include <stdio.h>
#endif
#include <gtk/gtk.h>

#include "globals.h"
#include "CElem.hpp"
#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "paint.h"

using namespace std;

extern vector<CElem*> elems;

/* pocet bodu, ktere byly naklikany 
 * urceno k indikaci, zda pri kliknuti vyvorit novy objekt, nebo pridat bod k poslednimu pridanemu objektu
 * ... nuluje se pri zmene typu mysiho kurzoru nebo pri dosazeni plneho poctu vrcholu daneho objektu */
int clicked_points=0;
/* souradnice posledniho stisknuti mysiho tlacitka. pri uvolneni se urci rozdil a rozhodne se, zda slo o
 * click nebo o drag */
double lx=-1, ly=-1;
/* workaround pro pripad, ze zprava mouse_move prijde driv nez mouse_down, ale behem jejiho vykonavani uzivatel stiskne mysitko */
int left_mouse_hold = 0;

void redraw_elements ()
{
	CElem::drawElems (elems);
}

/* prekresleni vseho */
void redraw ()
{
	draw_grid();
	redraw_elements();
	repaint();
}

void menu_new_clicked (GtkMenuItem *menuitem, gpointer user_data)
{
	vector<CElem*>::iterator it;
#ifdef DEBUG
	printf ("New file\n");
#endif
	/* TODO: zeptat se uzivatele jestli chce zahodit zmeny v predchozim dokumentu */
	it = elems.begin();
	while (it != elems.end()) {
		delete (*it);
		it++;
	}
	elems.clear();
	clicked_points = 0;
	redraw();
}


void
menu_quit_clicked                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	/* TODO: zeptej se na ulozeni rozpracovaneho dokumentu */
	gtk_main_quit ();
}


void
menu_cut_clicked                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
menu_copy_clicked                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
menu_paste_clicked                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
menu_delete_clicked                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	int number;
	/* zrusit nehotove objekty a nastavit kurzor na POINTER_SELECT */
	check_elem_completion();
	tool_toggled (NULL, (gpointer)POINTER_SELECT);
	number = CElem::delElems (elems);
#ifdef DEBUG
	printf ("%d elements deleted\n", number);
#endif
	redraw();
}


void
menu_about_clicked                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
}

/* zmena hodnoty snap_to_grid */
void tool_grid_toggled (GtkToggleToolButton *togglebutton, gpointer user_data)
{
	if (gtk_toggle_tool_button_get_active (togglebutton)) snap_to_grid=1;
	else snap_to_grid=0;
}

/* signal vymacknuti / zamacknuti nektereho zamackavaciho tlacitka */
void tool_toggled (GtkToolButton *toolbutton, gpointer user_data)
{
	/* vsechna tlacitka budou vymacknuta, jen to, na kterem nastal signal bude zamacknuto */
	int i;
	/* otestovat, jestli predchozi objekt byl vlozen kompletni */
	check_elem_completion();
	redraw();
	
	pointer = (gint) user_data;
#ifdef DEBUG
	printf ("tool %d toggles\n", pointer);
#endif
	for (i = 0; i < N_TOGGLE_BUTTONS; i++) {
		if (i == POINTER_ARC || i == POINTER_BEZLINE || i == POINTER_TEXT) continue;
		/* nejprve je potreba zakazat signaly od daneho tlacitka, jinak bychom se dostali do nekonecne smycky */
		g_signal_handler_block ((gpointer) togbut[i], toghandlers[i]);
		if (pointer == i)
			gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) togbut[i], TRUE);
		else
			gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) togbut[i], FALSE);
		/* znovupovoleni signalu */
		g_signal_handler_unblock ((gpointer) togbut[i], toghandlers[i]);
	}

	clicked_points = 0;
}


void tool_group_clicked (GtkToolButton* toolbutton, gpointer user_data)
{
	if (((int)user_data) == 0) {
		CElem::groupElems (elems);
	} else {
		CElem::ungroupElems (elems);
	}
	redraw ();
}

gboolean
canvas_expose                          (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	/* je potreba prekreslit cast kreslici plochy */
	gdk_draw_pixmap (widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)], bbuf,
			event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
	return FALSE;
}


gboolean canvas_configure (GtkWidget *widget, GdkEventConfigure *event, gpointer user_data)
{
	/* inicializace kreslici plochy
	 * pokazde, kdyz se meni velikost okna
	 * pokud "backbuffer" uz existuje, znicime ho */
	if (bbuf) {
		g_object_unref (bbuf);
	}
	bbuf = gdk_pixmap_new (widget->window, widget->allocation.width, widget->allocation.height, -1);

	/* premazani a vykresleni mrizky */
	can_width = widget->allocation.width;
	can_height = widget->allocation.height;
	draw_grid ();

	/* vykreslit jiz existujici graficke objekty */
	redraw();
	
	return TRUE;
}


gboolean canvas_mouse_down (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	if (event->button != 1) return FALSE;		/* zajimame se jen o leve tlacitko mysi */
	left_mouse_hold=1;
	switch (pointer) {
	case POINTER_SELECT:
		picture_to_world ((int)event->x, (int)event->y, &lx, &ly);
		break;
	case POINTER_MOVE:
		picture_to_world ((int)event->x, (int)event->y, &lx, &ly);
		if (snap_to_grid) {
			snapToGrid (lx, ly, &lx, &ly);
		}
		break;
	default:
		
		break;
	}
	
	return TRUE;
}


gboolean canvas_mouse_up (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	CElem *nel;		/* novy graficky element */
	vector<CElem*>::iterator it;
	double ptx, pty;	/* souradnice mysiho kurzoru prepoctene do logickych souradnic */
	double old_px, old_py;
	GdkColor col_fg, col_bg;
	bool filled;
	bool dashed;
	bool whole;
	int thick;
	double swap;
	GdkModifierType state;		/* stav mysich tlacitek a modifikatoru */
	int mx, my;
	int number;
	
	if (event->button == 3) {
		/* prave mysitko */
		/* ukonci vkladani objektu, ktery ma neznamy pocet bodu (multicara, polygon)
		 * u ostatnich nema vliv */
		check_elem_completion ();
		redraw ();
		return TRUE;
	}
	
	if (event->button != 1) return FALSE;		/* dale se zajimame jen o leve tlacitko mysi */
	left_mouse_hold = 0;
	picture_to_world ((int)event->x, (int)event->y, &ptx, &pty);
	if (snap_to_grid) {
		snapToGrid (ptx, pty, &ptx, &pty);
	}
	clicked_points++;

	/* aktualni barvy z color pickeru */
	gtk_color_button_get_color ((GtkColorButton*)pick_color_fg, &col_fg);
	gtk_color_button_get_color ((GtkColorButton*)pick_color_bg, &col_bg);
	/* vyplnovat nebo nevyplnovat */
	filled = (bool) gtk_toggle_tool_button_get_active ((GtkToggleToolButton*)toolbutton_filled);
	thick = (int) gtk_spin_button_get_value ((GtkSpinButton*)spin_width);
	dashed = (bool) gtk_toggle_tool_button_get_active ((GtkToggleToolButton*)toolbutton_dashed);
	whole = (bool) gtk_toggle_tool_button_get_active ((GtkToggleToolButton*)toolbutton_whole);
	
	switch (pointer) {
	case POINTER_SELECT:
		/* porovnat pt? a l? a poznat, jestli se jednalo o kliknuti nebo o vyber obdelnikove oblasti */
		gdk_window_get_pointer (event->window, &mx, &my, &state);
		picture_to_world ((int)event->x, (int)event->y, &ptx, &pty);
		swap = (ptx-lx) * (ptx-lx) + (pty-ly) * (pty-ly);
		if (swap < THRESHOLD_CLICK_QUADR) {
			/* mouse_down a mouse_up povazujeme za click */
			number = CElem::selectElems (elems, ptx, pty, (bool) (state & GDK_SHIFT_MASK), THRESHOLD_SELECT_QUADR);
#ifdef DEBUG
			printf ("%d elems selected by click\n", number);
#endif
		} else {
			/* byl proveden drag */
			number = CElem::selectElemsInside (elems, lx, ly, ptx, pty, (bool) (state & GDK_SHIFT_MASK), whole);
#ifdef DEBUG
			printf ("%d elems selected by rectangle\n", number);
#endif
		}
		lx = ly = -1;
		break;
	case POINTER_MOVE:
		lx = ly = -1;
		break;
	case POINTER_LINE:
		/* objekty zadavane dvema body */
		clicked_points %= 2;
		if (clicked_points == 1) {
			unselect_all();
			/* prvni bod cary - vytvorit novou caru */
			nel = new CElemLines (0);
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		}
		it = elems.end();
		it--;
		(*it)->movePoint (0, ptx, pty);
		break;
	case POINTER_MULTILINE:
		if (clicked_points == 1) {
			unselect_all();
			/* prvni bod multicary - vytvorit novou multicaru */
			nel = new CElemLines (0);
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		}
		it = elems.end();
		it--;
		(*it)->movePoint (0, ptx, pty);
		(*it)->addPoint (0, ptx, pty);
		break;
	case POINTER_TRIANGLE:
		clicked_points %= 3;
		if (clicked_points == 1) {
			unselect_all();
			/* prvni bod trojuhelniku */
			nel = new CElemPolygon (0);
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.colorRbg = col_bg.red;
			nel->common.colorGbg = col_bg.green;
			nel->common.colorBbg = col_bg.blue;
			nel->common.filled=filled;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		}
		it = elems.end();
		it--;
		(*it)->movePoint (0, ptx, pty);
		if (clicked_points == 2) {
			(*it)->addPoint (0, ptx, pty);
		}
		break;
	case POINTER_RECTANGLE:
		clicked_points %= 2;
		if (clicked_points == 1) {
			unselect_all();
			/* prvni bod obdelniku */
			nel = new CElemPolygon (0);
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.colorRbg = col_bg.red;
			nel->common.colorGbg = col_bg.green;
			nel->common.colorBbg = col_bg.blue;
			nel->common.filled=filled;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		} else {
			/* druhy bod urcujici opacny roh obdelnika */
			it = elems.end();
			it--;
			if ( ! (*it)->getPoint (3, &old_px, &old_py) ) {
				/* chyba - zadavame druhy bod obdelnika, ale prvni nam chybi */
				clicked_points = 0;
				elems.pop_back();
				break;
			}
			(*it)->movePoint (2, ptx, old_py);
			(*it)->movePoint (1, ptx, pty);
			(*it)->movePoint (0, old_px, pty);
		}
		break;
	case POINTER_ELLIPSE:
		clicked_points %= 2;
		if (clicked_points == 1) {
			unselect_all();
			/* stred elipsy */
			nel = new CElemEllipse ();
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.colorRbg = col_bg.red;
			nel->common.colorGbg = col_bg.green;
			nel->common.colorBbg = col_bg.blue;
			nel->common.filled=filled;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		} else {
			/* souradnice praveho dolniho rohu (prakticky jakehokoliv rohu) pravouhle obalky */
			it = elems.end();
			it--;
			(*it)->movePoint (1, ptx, pty);
		}
		break;
	case POINTER_ARC:
		break;
	case POINTER_POLYGON:
		if (clicked_points == 1) {
			unselect_all();
			/* pocatek noveho polygonu */
			nel = new CElemPolygon (0);
			if (nel == NULL) error (ERROR_NEW, __FILE__, __LINE__);
			nel->common.colorR = col_fg.red;
			nel->common.colorG = col_fg.green;
			nel->common.colorB = col_fg.blue;
			nel->common.colorRbg = col_bg.red;
			nel->common.colorGbg = col_bg.green;
			nel->common.colorBbg = col_bg.blue;
			nel->common.filled=filled;
			nel->common.selected = true;
			nel->common.lineWidth = thick;
			nel->common.lineStyle = dashed ? DASH_DASHED : DASH_SOLID;
			nel->addPoint (0, ptx, pty);
			elems.push_back (nel);
		}
		it = elems.end();
		it--;
		(*it)->movePoint (0, ptx, pty);
		(*it)->addPoint (0, ptx, pty);
		break;
	case POINTER_BEZLINE:
		break;
	case POINTER_TEXT:
		break;
	}

	redraw();
	return TRUE;
}

gboolean canvas_mouse_move (GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
	/* global min_points[] */
	double ptx,pty;		/* aktualni souradnice ve svetovych souradnicich */
	vector<CElem*>::iterator it;
	double ox, oy;		/* souradnice jinych nez poslednich bodu elementu (pro obdelnik) */
	int mx, my;		/* souradnice pozice mysi v pixelech */
	GdkModifierType state;		/* stav mysich tlacitek a modifikatoru */
	struct point_s pts[4];
	
	/* pouzivame pointer_motion_hint, takze se musime zeptat gdk na mysi souradnice explicitne */
	gdk_window_get_pointer (event->window, &mx, &my, &state);
	picture_to_world (mx, my, &ptx, &pty);
	
	if (pointer != POINTER_SELECT && pointer != POINTER_MOVE) {
		if (snap_to_grid) {
			snapToGrid (ptx, pty, &ptx, &pty);
		}
		if ((clicked_points > 0) /*&& (clicked_points < min_points[pointer])*/) {
			it = elems.end();
			it--;
			if (pointer == POINTER_ELLIPSE) {
				(*it)->movePoint (1, ptx, pty);
			} else if (pointer == POINTER_RECTANGLE) {
				if (!(*it)->getPoint (3, &ox, &oy)) return FALSE;
				(*it)->movePoint (2, ptx, oy);
				(*it)->movePoint (1, ptx, pty);
				(*it)->movePoint (0, ox, pty);
			} else {
				(*it)->movePoint (0, ptx, pty);
			}
			redraw();
		}
	} else if (pointer == POINTER_SELECT) {
		if (state & GDK_BUTTON1_MASK && left_mouse_hold) {
			/* je stisknute leve mysitko */
			redraw ();
			pts[0].x = lx;  pts[0].y = ly;
			pts[1].x = ptx; pts[1].y = ly;
			pts[2].x = ptx; pts[2].y = pty;
			pts[3].x = lx;  pts[3].y = pty;
			draw_polygon (pts, 4, 0,0,0, 15000,15000,15000, 1, DASH_DASHED, 0);
		}
	} else if (pointer == POINTER_MOVE) {
		if (state & GDK_BUTTON1_MASK) {
			if (snap_to_grid) {
				snapToGrid (ptx, pty, &ptx, &pty);
			}
			CElem::moveElems (elems, ptx - lx, pty-ly);
			lx = ptx;
			ly = pty;
			redraw();
		}
	}

	return TRUE;
}

/* zmena vybrane barvy (gpointer==0 .. fg, gpointer==1 .. bg) - prave vybranym objektum zmenit barvu */
void tool_color_set (GtkColorButton *button, gpointer user_data)
{
	GdkColor col;
	vector<CElem*>::iterator it;
	
	if ((int) user_data == 0)
		gtk_color_button_get_color ((GtkColorButton*)pick_color_fg, &col);
	else
		gtk_color_button_get_color ((GtkColorButton*)pick_color_bg, &col);
	
	it = elems.begin();
	while (it != elems.end()) {
		if ((*it)->common.selected) {
			if ((int) user_data == 0) {
				(*it)->common.colorR = col.red;
				(*it)->common.colorG = col.green;
				(*it)->common.colorB = col.blue;
			} else {
				(*it)->common.colorRbg = col.red;
				(*it)->common.colorGbg = col.green;
				(*it)->common.colorBbg = col.blue;
			}
		}
		it++;
	}
	redraw ();
}

/* zmena vybraneho fontu */
void tool_font_set (GtkColorButton *button, gpointer user_data)
{
}

/* reakce na stisknuti nektereho z tlacitek zoomu */
void zoom_clicked (GtkToolButton* button, gpointer user_data)
{
	/* TODO: upravit podle zoomu rozmezi posuvniku */
/*	gdouble a,b;
	GValue va,vb;

	g_object_get_property (hsa, "value", &va);
	g_object_get_property (vsa, "value", &vb);*/
	switch ((int) user_data) {
	case 0:	
/*		hsa->value /= zoom;
		vsa->value /= zoom;*/
		zoom = 1.0;
		break;
	case 1:
/*		hsa->value /= 2.0;
		vsa->value /= 2.0;*/
		zoom /= 2.0;
		break;
	case 2:
		zoom *= 2.0;
		break;
	}
	
	/* zasedivime tlacitka zoomu, pokud jsme uz moc zazoomovani */
	check_zoom_gray();
	
	/* zmena rozsahu posuvniku */
/*	hsa->lower = vsa->lower = RANGE_MIN * zoom;
	hsa->upper = vsa->upper = RANGE_MAX * zoom;
	if ((int)user_data == 2) {
		hsa->value *= 2.0;
		vsa->value *= 2.0;
	}*/
	
	redraw ();
}

/* oznacenym objektum nastavit nebo zrusit atribut filled */
void toggle_filled (gboolean active)
{
	vector<CElem*>::iterator it;
	
	it = elems.begin();
	while (it != elems.end()) {
		if ((*it)->common.selected) {
			(*it)->common.filled = (bool) active;
		}
		it++;
	}
	redraw ();
}

/* oznacenym objektum nastavit nebo zrusit atribut linestyle */
void toggle_dashed (gboolean active)
{
	vector<CElem*>::iterator it;
	it = elems.begin();
	while (it != elems.end()) {
#ifdef DEBUG
		printf ("pointer: %p\n", (*it));
#endif	
		if ((*it)->common.selected) {
			(*it)->common.lineStyle = (active)? DASH_DASHED : DASH_SOLID;
		}
		it++;
	}
	redraw ();
}

/* oznacenym objektum nastavit tloustku cary */
void spin_width_change (GtkSpinButton *spin, gpointer user_data)
{
	int thick;
	vector<CElem*>::iterator it;
	thick = (int) gtk_spin_button_get_value (spin);
	
	it = elems.begin();
	while (it != elems.end()) {
		if ((*it)->common.selected) {
			(*it)->common.lineWidth = thick;
		}
		it++;
	}
	redraw ();
}

/* zmena hodnoty nektereho posuvniku */
gboolean scroll_change (GtkRange *range, GtkScrollType scr, gdouble value, gpointer user_data)
{
	int i;
	if ((int)user_data == 0) i = 0; else i = 1;
	scroll[i] = value;
	redraw();
	return FALSE;
}

/* funkce testuje, jestli posledni zadany objekt ma dostatecny pocet vrcholu a pokud ne, tak ho odebere */
void check_elem_completion ()
{
	/* global min_points[] */
	vector<CElem*>::iterator it;
	
	if (clicked_points > 0 && clicked_points < min_points[pointer]) {
		/* odstranit posledni element */
		elems.pop_back();
	}
	if (((pointer == POINTER_MULTILINE) || (pointer == POINTER_POLYGON)) && (clicked_points >= min_points[pointer])) {
		/* u polygonu a multicary musime odebrat posledni bod, ktery uz uzivatel nepotvrdil kliknutim */
		it = elems.end();
		it--;
		(*it)->delPoint (0);
	}
	
	clicked_points = 0;
}

/* zrusi priznak "vybran" vsem elementum */
void unselect_all ()
{
	vector<CElem*>::iterator it;
	it = elems.begin();
	while (it != elems.end()) {
		(*it)->common.selected = false;
		it++;
	}
}

