#include <stdlib.h>
#include <gtk/gtk.h>
#include <math.h>

#ifdef DEBUG
#include <stdio.h>
#endif

#include "globals.h"
#include "interface.h"
#include "paint.h"
#include "algebra.h"

/* prepocet logickych souradnic na souradnice v bitmape. bere v potaz zoom i posunuti scrollbaru */
void world_to_picture (double in_x, double in_y, int *out_x, int *out_y)
{
	*out_x = (int)((in_x - scroll[0]) * zoom);
	*out_y = (int)((in_y - scroll[1]) * zoom);
	return;
}

void picture_to_world (int in_x, int in_y, double *out_x, double *out_y)
{
	*out_x = ((double)in_x) / zoom + scroll[0];
	*out_y = ((double)in_y) / zoom + scroll[1];
}

void snapToGrid (double in_x, double in_y, double *out_x, double *out_y)
{
	*out_x = roundto( in_x, GRID_SPAN );
	*out_y = roundto( in_y, GRID_SPAN );
	return;
}

int zoom_thickness (int in)
{
	int res = (int) (in * zoom);
	if (res < 1) res = 1;
	return res;
}

/* prekresleni okna */
void repaint ()
{
	gtk_widget_queue_draw (canvas);
}

/* vymazani bufferu a nakresleni mrizky */
void draw_grid ()
{
	GdkGC *grid_gc;
	GdkGC *grid2_gc;
	GdkColor color;
	gint x,y,i;
	gint delta = GRID_SPAN * zoom;
	gint off, ffo;

	grid_gc = gdk_gc_new (bbuf);
	color.red = color.green = color.blue = 65535;
	gdk_gc_set_rgb_fg_color (grid_gc, &color);
	grid2_gc = gdk_gc_new (bbuf);
	/* vymazani barvou pozadi */
	gdk_draw_rectangle (bbuf, grid_gc, TRUE, 0,0, can_width, can_height);

	/* nakreslit mriz */
	gdk_gc_set_line_attributes (grid_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
	gdk_gc_set_line_attributes (grid2_gc, 2, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
	color.red = color.green = color.blue = 50000;
	gdk_gc_set_rgb_fg_color (grid_gc, &color);
	gdk_gc_set_rgb_fg_color (grid2_gc, &color);
	
	off = - ((int)(scroll[0] * zoom) % delta);
	ffo = ((int) (scroll[0] * zoom / delta)) % 5;
	for (x=off, i=ffo; x < can_width; x += delta, i++) {
		gdk_draw_line (bbuf, (i%5)? grid_gc : grid2_gc, x, 0, x, can_height);
	}
	off = - ((int)(scroll[1] * zoom) % delta);
	ffo = ((int) (scroll[1] * zoom / delta)) % 5;
	for (y=off, i=ffo; y < can_height; y += delta, i++) {
		gdk_draw_line (bbuf, (i%5)? grid_gc : grid2_gc, 0, y, can_width, y);
	}

	g_object_unref (grid_gc);
}

/* nakresleni zvyraznenych bodu */
void draw_points (struct point_s pts[], int pt_count, unsigned short r, unsigned short g, unsigned short b)
{
	int i, x, y;
	GdkGC *ptgc;
	GdkGC *border;
	GdkColor color;
	int tmp;
	tmp = snap_to_grid;
	snap_to_grid = FALSE;

	ptgc = gdk_gc_new (bbuf);
	border = gdk_gc_new (bbuf);
	gdk_gc_set_line_attributes (ptgc, 1, GDK_LINE_SOLID, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	gdk_gc_set_line_attributes (border, 1, GDK_LINE_SOLID, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	color.red = r;
	color.green = g;
	color.blue = b;
	gdk_gc_set_rgb_fg_color (ptgc, &color);
	color.red = color.green = color.blue = 0;
	gdk_gc_set_rgb_fg_color (border, &color);
	
	for (i=0; i<pt_count; i++) {
		world_to_picture (pts[i].x, pts[i].y, &x, &y);
		gdk_draw_rectangle (bbuf, ptgc, TRUE, x-PAINT_PT_WIDTH, y-PAINT_PT_WIDTH, 2*PAINT_PT_WIDTH, 2*PAINT_PT_WIDTH);
		gdk_draw_rectangle (bbuf, border, FALSE, x-PAINT_PT_WIDTH-1, y-PAINT_PT_WIDTH-1, 2*PAINT_PT_WIDTH+1, 2*PAINT_PT_WIDTH+1);
	}
	
	snap_to_grid = tmp;
	g_object_unref (ptgc);
	g_object_unref (border);
}

/* nakresleni cary */
void draw_line (double x1, double y1, double x2, double y2, unsigned short r, unsigned short g, unsigned short b, int thickness, int dash)
{
	int p1x, p1y, p2x, p2y;
	int dash_style;
	GdkGC *line_gc;
	GdkColor color;
	int tmp;

	tmp = snap_to_grid;
	snap_to_grid = FALSE;
	
	world_to_picture (x1, y1, &p1x, &p1y);
	world_to_picture (x2, y2, &p2x, &p2y);
	thickness = zoom_thickness (thickness);
	line_gc = gdk_gc_new (bbuf);
	dash_style = (dash == DASH_DASHED) ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID;
	gdk_gc_set_line_attributes (line_gc, thickness, dash_style, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	color.red = r;
	color.green = g;
	color.blue = b;
	gdk_gc_set_rgb_fg_color (line_gc, &color);
	gdk_draw_line (bbuf, line_gc, p1x, p1y, p2x, p2y);

	snap_to_grid = tmp;
	g_object_unref (line_gc);
}

/* nakresleni multicary */
void draw_multiline (struct point_s pts[], int pt_count, unsigned short r, unsigned short g, unsigned short b, int thickness, int dash)
{
	int p1x, p1y, p2x, p2y;
	int i, dash_style;
	GdkGC *line_gc;
	GdkColor color;
	int tmp;
	
	if (pt_count < 2) return;
	tmp = snap_to_grid;
	snap_to_grid = FALSE;
	world_to_picture (pts[0].x, pts[0].y, &p1x, &p1y);
	thickness = zoom_thickness (thickness);
	line_gc = gdk_gc_new (bbuf);
	dash_style = (dash == DASH_DASHED) ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID;
	gdk_gc_set_line_attributes (line_gc, thickness, dash_style, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	color.red = r;
	color.green = g;
	color.blue = b;
	gdk_gc_set_rgb_fg_color (line_gc, &color);
	for (i = 1; i < pt_count ; i++) {
		world_to_picture (pts[i].x, pts[i].y, &p2x, &p2y);
		gdk_draw_line (bbuf, line_gc, p1x, p1y, p2x, p2y);
		p1x = p2x;
		p1y = p2y;
	}

	snap_to_grid = tmp;
	g_object_unref (line_gc);
}

/* nakresleni polygonu, pole pts ma rozmer [pt_count][2], parametry thickness a dash se vztahuji na obvodovou linku polygonu */
void draw_polygon (struct point_s pts[], int pt_count, unsigned short bg_r, unsigned short bg_g, unsigned short bg_b, unsigned short border_r,
	       	unsigned short border_g, unsigned short border_b, int thickness, int dash, int filled)
{
	GdkPoint *_pt;
	int i, dash_style;
	GdkGC *poly_gc;
	GdkColor color;
	int tmp;
	
	thickness = zoom_thickness (thickness);
	if (pt_count < 3) return;
	tmp = snap_to_grid;
	snap_to_grid = FALSE;
	_pt = (GdkPoint*) malloc (pt_count * sizeof (GdkPoint));
	if (_pt == NULL) {
		error (ERROR_MALLOC, __FILE__, __LINE__);
		return;
	}
	
	for (i = 0; i < pt_count; i++) {
		world_to_picture (pts[i].x, pts[i].y, &(_pt[i].x), &(_pt[i].y));
	}
	
	poly_gc = gdk_gc_new (bbuf);
	if (filled) {
		/* u vyplneneho polygonu nejprve nakreslit pozadi */
		gdk_gc_set_fill (poly_gc, GDK_SOLID);
		color.red = bg_r;
		color.green = bg_g;
		color.blue = bg_b;
		gdk_gc_set_rgb_fg_color (poly_gc, &color);
		gdk_draw_polygon (bbuf, poly_gc, TRUE, _pt, pt_count);
	}
	/* vykreslit obrys */
	dash_style = (dash == DASH_DASHED) ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID;
	gdk_gc_set_line_attributes (poly_gc, thickness, dash_style, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	color.red = border_r;
	color.green = border_g;
	color.blue = border_b;
	gdk_gc_set_rgb_fg_color (poly_gc, &color);
	gdk_draw_polygon (bbuf, poly_gc, FALSE, _pt, pt_count);
	
	snap_to_grid = tmp;
	free (_pt);
	g_object_unref (poly_gc);
}

/* nakresleni eliptickeho oblouku, zadaneho stredem a horizontalnim a vertikalnim polomerem. uhly jsou ve stupnich */
void draw_arch (double cent_x, double cent_y, double rad_h, double rad_v, double angle_start, double angle_stop,
		unsigned short bg_r, unsigned short bg_g, unsigned short bg_b, unsigned short border_r, unsigned short border_g, unsigned short border_b,
		int thickness, int dash, int filled)
{
	int lt_x, lt_y, rb_x, rb_y;
	int swap;
	int dash_style;
	GdkGC *arc_gc;
	GdkColor color;
	int tmp;
	
	thickness = zoom_thickness (thickness);
	tmp = snap_to_grid;
	snap_to_grid = FALSE;
	world_to_picture (cent_x-rad_h, cent_y-rad_v, &lt_x, &lt_y);
	world_to_picture (cent_x+rad_h, cent_y+rad_v, &rb_x, &rb_y);
	if (lt_x > rb_x) {
		swap = lt_x;
		lt_x = rb_x;
		rb_x = swap;
	}
	if (lt_y > rb_y) {
		swap = lt_y;
		lt_y = rb_y;
		rb_y = swap;
	}
	angle_start *= 64;
	angle_stop *= 64;

	arc_gc = gdk_gc_new (bbuf);
	if (filled) {
		/* u vyplneneho oblouku/elipsy nejprve nakreslit pozadi */
		gdk_gc_set_fill (arc_gc, GDK_SOLID);
		color.red = bg_r;
		color.green = bg_g;
		color.blue = bg_b;
		gdk_gc_set_rgb_fg_color (arc_gc, &color);
		gdk_draw_arc (bbuf, arc_gc, TRUE, lt_x, lt_y, rb_x - lt_x, rb_y - lt_y, angle_start, angle_stop - angle_start);
	}
	/* vykreslit obrys */
	dash_style = (dash == DASH_DASHED) ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID;
	gdk_gc_set_line_attributes (arc_gc, thickness, dash_style, LINE_CAP_STYLE, LINE_JOIN_STYLE);
	color.red = border_r;
	color.green = border_g;
	color.blue = border_b;
	gdk_gc_set_rgb_fg_color (arc_gc, &color);
	gdk_draw_arc (bbuf, arc_gc, FALSE, lt_x, lt_y, rb_x - lt_x, rb_y - lt_y, angle_start, angle_stop - angle_start);

	snap_to_grid = tmp;
	g_object_unref (arc_gc);
}

/* nakresleni elipsy zadane stredem a horizontalnim a vertikalnim polomerem */
void draw_ellipse (double cent_x, double cent_y, double rad_h, double rad_v, unsigned short bg_r, unsigned short bg_g,
		unsigned short bg_b, unsigned short border_r, unsigned short border_g, unsigned short border_b, int thickness, int dash, int filled)
{
	draw_arch (cent_x, cent_y, rad_h, rad_v, 0, 360, bg_r, bg_g, bg_b, border_r, border_g, border_b, thickness, dash, filled);
}

/* napsani textu, text je C-style ASCIIZ */
void draw_text (double topleft_x, double topleft_y, char* text, int size, int weight, unsigned short r, unsigned short g, unsigned short b, char *family)
{
	PangoLayout *pl;
	PangoFontDescription *font;
	GdkGC* text_gc;
	GdkColor color;
	int _size;
	int tmp;

	weight = zoom_thickness (weight);
	if (weight < 7) weight = 7;
	tmp = snap_to_grid;
	snap_to_grid = FALSE;
	text_gc = gdk_gc_new (bbuf);
	color.red = r;
	color.green = g;
	color.blue = b;
	gdk_gc_set_rgb_fg_color (text_gc, &color);
	/* nastaveni pisma */
	pl = gtk_widget_create_pango_layout (canvas, text);
	font = pango_context_get_font_description (can_pang_con);
	pango_font_description_set_family (font, family);
	_size = pango_font_description_get_size (font);
#ifdef DEBUG
	printf ("font size %d\n", _size);
#endif
	pango_font_description_set_weight (font, weight);
	pango_font_description_set_size (font, size * _size);
	pango_layout_set_font_description (pl, font);
	gdk_draw_layout (bbuf, text_gc, topleft_x, topleft_y, pl);
	pango_font_description_set_size (font, _size);
	pango_layout_set_font_description (pl, font);

	snap_to_grid = tmp;
	g_object_unref (pl);
	g_object_unref (text_gc);
}

